// SPDX-FileCopyrightText: Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: EUPL-1.2

use super::Cleaner;
use crate::decision::{Action, Decision};
use anyhow::{Context, Result};
use async_trait::async_trait;
use std::path::PathBuf;
use tokio::sync::mpsc;

pub struct ListCleaner {
    list_type: ListType,
    list: PathBuf,
}

pub enum ListType {
    Allow,
    Block,
}

impl ListCleaner {
    pub fn new(list_type: ListType, list: PathBuf) -> Self {
        Self { list_type, list }
    }
}

#[async_trait]
impl Cleaner for ListCleaner {
    async fn run(
        &mut self,
        mut input: mpsc::Receiver<Decision>,
        output: mpsc::Sender<Decision>,
    ) -> Result<()> {
        let current_dir = std::env::current_dir()?;

        // We use a gitignore builder because it has the syntax and
        // expressivity we need. Checking if a file "is ignored"
        // will tell us if it matches against the allow-/block-list
        let mut matcher_builder = ignore::gitignore::GitignoreBuilder::new(current_dir);
        if let Some(err) = matcher_builder.add(&self.list) {
            anyhow::bail!("list: error parsing {}: {}", self.list.display(), err);
        }

        let matcher = matcher_builder
            .build()
            .with_context(|| format!("list: cannot build matcher from {}", self.list.display()))?;

        let (action_name, action) = match self.list_type {
            ListType::Allow => ("allow", Action::Keep),
            ListType::Block => ("block", Action::Remove),
        };

        while let Some(mut decision) = input.recv().await {
            if decision.action == Action::Undecided
                && matcher.matched(&decision.path, false).is_ignore()
            {
                log::info!("{}list: marking {}", action_name, decision.path.display());
                decision.action = action;
            }

            output.send(decision).await?;
        }

        Ok(())
    }
}
